import pickle
import os


import inspect

class key_memoized(object):
    def __init__(self, func):
       self.func = func
       self.cache = {}

    def __call__(self, *args, **kwargs):
        key = self.key(args, kwargs)
        if key not in self.cache:
            self.cache[key] = self.func(*args, **kwargs)
            print(self.cache)
        else:
            print('using cached value')
        return self.cache[key]

    def normalize_args(self, args, kwargs):
        spec = inspect.getargs(self.func.__code__).args
        return dict(**kwargs, **dict(zip(spec, args)))
        #return dict(kwargs.items() + zip(spec, args))

    def key(self, args, kwargs):
        a = self.normalize_args(args, kwargs)
        return tuple(sorted(a.items()))


def cached(func):
    """memorizer decorator function to cache temporary results.
    """
    cache = dict()

    def memoized_func(*args):
        if args in cache:
            return cache[args]
        result = func(*args)
        cache[args] = result
        return result

    return memoized_func


#def disk_cached(cachefile=False):
#    """memorizer decorator function to cache temporary results.
#    """
#    cache = dict()
#    if cachefile:
#        # load from file
#        if os.path.isfile(cachefile):
#            with open(cachefile, 'rb') as handle:
#                cache = pickle.load(handle)
#
#    def wrapper(func):
#
#        def memoized_func(*args):
#            if args in cache:
#                return cache[args]
#            result = func(*args)
#            cache[args] = result
#            if cachefile:
#                with open(cachefile, 'wb') as handle:
#                    pickle.dump(your_data, handle, protocol=pickle.HIGHEST_PROTOCOL)
#            return result
#        return memoized_func
#
#    return memorized_func








from cdo import Cdo
from .cache_deco import cached
import pandas as pd

import logging

class Cmorizer():

    codemap   = {130: 'ta', 134: 'ps', 131: 'ua', 132: 'va',
                 133: 'hus',  34 : 'tos',  31 :'sic',  129 : 'orog',
                 172: 'sftlf', 139: 'tsl1', 170: 'tsl2', 183: 'tsl3',
                 238: 'tsn', 236: 'tsl4', 246: 'clw', 141: 'snw', 198: 'src' }

    def __init__(self):
        self.cdo = Cdo(logging=True)

    @cached
    def _to_regular(self, filename):
        """converts ecmwf spectral grib data to regular gaussian netcdf.

        cdo is used to convert ecmwf grid data to netcdf depending on the gridtype:
        For 'gaussian_reduced': cdo setgridtype,regular
            'spectral'        : cdo sp2gpl

        This follows the recommendation from the ECMWF Era5 Documentation.
        We also invert the latitudes to stick with cmor standard.

        """
        gridtype = self._gridtype(filename)
        if gridtype == 'gaussian_reduced':
            #return self.cdo.copy(options='-R -f nc', input=filename)
            gaussian = self.cdo.setgridtype('regular', options='-f nc', input=filename)
        elif gridtype == 'spectral':
            gaussian = self.cdo.sp2gpl(options='-f nc', input=filename)
        elif gridtype == 'gaussian':
            gaussian =  self.cdo.copy(options='-f nc', input=filename)
        else:
            raise Exception('unknown grid type for conversion to regular grid: {}'.format(gridtype))
        return self.cdo.invertlat(input=gaussian)

    @cached
    def _nc_convert(self, filename):
        regular = self._to_regular(filename)
        code = self._get_code(regular)
        if code in self.codemap:
            return self.cdo.setname(self.codemap[code], input=regular)
        else:
            return regular

    def _gridtype(self, filename):
        griddes = self.cdo.griddes(input=filename)
        griddes = {entry.split('=')[0].strip():entry.split('=')[1].strip() for entry in griddes if '=' in entry}
        return griddes['gridtype']

    def _get_code(self, filename):
        showcode = self.cdo.showcode(input=filename)
        print(showcode)
        return showcode



class FileManager():

    date_fmt = "%Y-%m-%dT%H:%M:%S"

    def __init__(self, df):
        self.df = df
        self.df['startdate'] = pd.to_datetime(self.df['startdate'])
        self.df['enddate'] = pd.to_datetime(self.df['enddate'])
        self.cmor = Cmorizer()
        self.cdo = Cdo(logging=True)

    @property
    def file_list(self):
        return self.df['path'].values

    @property
    def code(self):
        """returns code from dataframe, should be unique
        """
        code = self.df['code'].unique()
        if len(code) == 1:
            return code[0]
        else:
            raise Exception('code is not unique in file list')

    @cached
    def tmp_file_by_date(self, datetime, nc=True):
        filename = self.get_file_by_date(datetime)
        date_str = datetime.strftime(self.date_fmt)
        logging.debug('selecting {} from {}'.format(date_str, filename))
        if nc:
            return self.cmor._nc_convert( self.cdo.seldate(date_str, input=filename ) )
        else:
            return self.cdo.seldate(date_str, input=filename )

    def get_file_by_date(self, datetime):
        """search for a file in the dataframe that should contain the datetime.
        """
        df = self.df[ (datetime >= self.df['startdate']) & (datetime <= self.df['enddate']) ]
        if len(df['path'].values) == 0:
            raise Exception('no files found for {}'.format(datetime))
        elif len(df['path'].values) > 1:
            raise Exception('date selection must be unique')
        else:
            return df['path'].values[0]

    def _ref_file(self, filename=None, nc=True):
        """the reference dataset is used to access dataset attributes.

        This is required to give the preprocessor acces to, e.g., grid
        and calendar information.
        """
        if filename is None:
            filename = self.file_list[0]
        logging.info('getting reference dataset from {}'.format(filename))
        if nc:
            return self._nc_convert( self._reftimestep( filename ) )
        else:
            return self._reftimestep( filename )




import os

import datetime as dt
from cdo import Cdo
import logging
from netCDF4 import Dataset, MFDataset, date2index, num2date, date2num
import xarray as xr
import pandas as pd
from .cache_deco import cached, key_memoized
import numpy as np

from .exp import ExpVars

from . import filemanager as fm

cdo_logging = True

ak_valid = ['ap_bnds','a_bnds']
bk_valid = ['b_bnds']




class NC4Dataset():

    def __init__(self, file_list=None, ds=None, time_axis='time', **kwargs):
        self.time_axis = time_axis
        if file_list and not isinstance(file_list, list):
            file_list = [file_list]
        self.file_list = file_list
        if ds is None:
            if len(self.file_list) == 1:
                self.ds = Dataset(self.file_list[0], **kwargs)
            elif len(self.file_list) > 1:
                self.ds = MFDataset(self.file_list, **kwargs)
        else:
            self.ds = ds
        if file_list is None and ds is None:
            logging.warning('empty file list or dataset!')

    def __str__(self):
        return str(self.ds)

#    def __getattr__(self, item):
#        logging.debug('getting {} from dataset...'.format(item))
#        return getattr(self.ds, item)

#    @property
#    def file_list(self):
#        return list(self.df['path'])

    @property
    def calendar(self):
        return self.ds.variables[self.time_axis].calendar

    @property
    def units(self):
        return self.ds.variables[self.time_axis].units

    @property
    def variables(self):
        return self.ds.variables

    @property
    def positive_down(self):
        return self.ds.variables['lev'].positive == "down"

    def ncattrs(self):
        return self.ds.ncattrs()

    def global_coordinates(self, varname=None):
        return fm.get_lonlat_2d(self.ds)


    def get_vc(self):
        """Reads the vertical hybrid coordinate from a dataset.
        """
        ak_bnds = None
        bk_bnds = None
        for ak_name in ak_valid:
            if ak_name in self.ds.variables:
                ak_bnds = self.ds.variables[ak_name]
                logging.info('using {} for akgm'.format(ak_name))
        for bk_name in bk_valid:
            if bk_name in self.ds.variables:
                bk_bnds = self.ds.variables[bk_name]
                logging.info('using {} for bkgm'.format(bk_name))
        if not all([ak_bnds, bk_bnds]):
            logging.error('could not identify vertical coordinate, tried: {}, {}'.format(ak_valid, bk_valid))
            raise Exception('incomplete input dataset')
#        ak_bnds, bk_bnds  = (ak_bnds[:1], bk_bnds[:,1])
        nlev = ak_bnds.shape[0]
        ak = np.zeros([nlev+1], dtype=np.float64)
        bk = np.ones([nlev+1], dtype=np.float64)
        #ak[1:] = (ak_bnds[:])
        #bk[1:] = (bk_bnds[:])
        if self.positive_down:
            #ak[:-1] = np.flip(ak_bnds[:])
            #bk[:-1] = np.flip(bk_bnds[:])
            ak[:-1] = np.flip(ak_bnds[:,1])
            bk[:-1] = np.flip(bk_bnds[:,1])
        else:
            ak[1:] = np.flip(ak_bnds[:,1])
            bk[1:] = np.flip(bk_bnds[:,1])
        return ak, bk

    #def ncattrs_dict(self, varname=None):
    #    if varname:
    #        return {attr:self.ds.variables[varname].getncattr(attr)
    #                for attr in self.ds.variables[varname].ncattrs()}
    #    else:
    #        return {attr:self.ds.getncattr(attr)
    #                for attr in self.ds.ncattrs()}

    def ncattrs_dict(self, varname=None):
        if varname:
            return {attr:getattr(self.ds.variables[varname], attr)
                    for attr in self.ds.variables[varname].ncattrs()}
        else:
            return {attr:getattr(self.ds, attr)
                    for attr in self.ds.ncattrs()}

    def get_index_by_date(self, dates):
        logging.debug('dates: {}'.format(dates))
        return date2index(dates,
                self.ds.variables[self.time_axis], select='exact')

    def get_date_by_num(self, nums):
        return num2date(nums, self.units, calendar=self.calendar)

    def get_num_by_date(self, dates):
        return date2num(dates, self.units, calendar=self.calendar)

    def get_num_by_index(self, index):
        return self.variables[self.time_axis][index]

    def get_date_by_index(self, index):
        num = self.ds.variables[self.time_axis][index]
        return self.get_date_by_num(num)

    def get_timestep(self, date, varname):
        ix = self.get_index_by_date(date)
        return self.ds.variables[varname][ix]

    def get_variable(self, varname):
        return self.ds.variables[varname]

    def dynamic(self, varname):
        return self.ds.variables[varname].dimensions[0] == self.time_axis

    def data_by_date(self, variable, date):
        return self.timestep(variable, self.get_index_by_date(date))

    def static(self, varname):
        return not self.dynamic(varname)

    def timestep(self, varname, timestep=0):
        if self.dynamic(varname):
            return self.variables[varname][timestep]
        else:
            return self.variables[varname][:]

    def threeD(self, varname):
        return len(self.timestep(varname).shape) == 3

    def twoD(self, varname):
        return not self.threeD(varname)

    def nlev(self, varname=None):
        if self.threeD(varname):
            return self.ds.variables[varname].shape[1:-2][0]
        else:
            return 1

    def to_netcdf(self, varname=None, timestep=None, destination=None):
        ds = copy_dataset(self.ds, varname=varname, timestep=timestep, destination=destination)
        filepath = ds.filepath()
        ds.close()
        return filepath




#class NC4MFDataset(NC4Dataset):
#
#    def __init__(self, df, time_axis='time', **kwargs):
#        NC4Dataset.__init__(self, df, time_axis, **kwargs)
#
#    def ncattrs_dict(self, varname=None):
#        if varname:
#            return {attr:getattr(self.ds.variables[varname], attr)
#                    for attr in self.ds.variables[varname].ncattrs()}
#        else:
#            return {attr:getattr(self.ds, attr)
#                    for attr in self.ds.ncattrs()}
#
#    def getncattr(self, attr):
#        return getattr(self.ds, attr)
#


class XRDataset():

    def __init__(self, file_list, time_axis='time'):
        self.time_axis = time_axis
        self.ds = xr.open_dataset(file_list, decode_times=False, decode_cf=False,
                decode_coords=False, chunks={'time':1, 'lat':1000, 'lon':1000})


class XRMFDataset():

    def __init__(self, file_list, time_axis='time'):
        self.time_axis = time_axis
        self.ds = xr.open_mfdataset(file_list)



class ECMWF(NC4Dataset):

    date_fmt = "%Y-%m-%dT%H:%M:%S"
    codemap   = {130: 'ta', 134: 'ps', 131: 'ua', 132: 'va',
                 133: 'hus',  34 : 'tos',  31 :'sic',  129 : 'orog',
                 172: 'sftlf', 139: 'tsl1', 170: 'tsl2', 183: 'tsl3',
                 238: 'tsn', 236: 'tsl4', 246: 'clw', 141: 'snw', 198: 'src', 235: 'skt' ,
                 39: 'swvl1', 40: 'swvl2', 41: 'swvl2'}


    def __init__(self,df=None, time_axis='time'):
        self.cdo = Cdo(logging=True, tempdir=os.path.join(ExpVars.scratch,'python-cdo'))
        self.time_axis = time_axis
        self.df = df
        self.df['startdate'] = pd.to_datetime(self.df['startdate'])
        self.df['enddate'] = pd.to_datetime(self.df['enddate'])
        self.tmp_ds = {}

    @property
    def file_list(self):
        return self.df['path'].values

    def threeD(self, varname):
        """soil fields are handled as 2D although they have a depth dimension.
        """
        return len(self.timestep(varname).shape) == 3 and self.timestep(varname).shape[0] > 1

    @property
    def positive_down(self):
        """ERA5 data has the correct layer numbering for REMO.
        """
        return False

    @property
    def code(self):
        """returns code from dataframe, should be unique
        """
        code = self.df['code'].unique()
        if len(code) == 1:
            return code[0]
        else:
            raise Exception('code is not unique in file list')

    @property
    def ds(self):
        return Dataset(self._ref_file())

    def get_vc(self):
        """Reads the vertical hybrid coordinate from a dataset.
        """
        return self.variables['hyai'], self.variables['hybi']

    @cached
    def _tmp_file_by_date(self, datetime, nc=True):
        filename = self.get_file_by_date(datetime)
        date_str = datetime.strftime(self.date_fmt)
        logging.debug('selecting {} from {}'.format(date_str, filename))
        if nc:
            return self._nc_convert( self.cdo.seldate(date_str, input=filename ) )
        else:
            return self.cdo.seldate(date_str, input=filename )

    def _ref_file(self, filename=None, nc=True):
        """the reference dataset is used to access dataset attributes.

        This is required to give the preprocessor acces to, e.g., grid
        and calendar information.
        """
        if filename is None:
            filename = self.file_list[0]
        logging.info('getting reference dataset from {}'.format(filename))
        if nc:
            return self._nc_convert( self._reftimestep( filename ) )
        else:
            return self._reftimestep( filename )

        #logging.debug('reference dataset: {}'.format(self.ref_ds.filepath()))
    @cached
    def _reftimestep(self, filename):
        return self.cdo.seltimestep( 1, input=filename)

    @cached
    def _nc_convert(self, filename):
        regular = self._to_regular(filename)
        if self.code in self.codemap:
            return self.cdo.setname(self.codemap[self.code], input=regular)
        else:
            return regular

    def _gridtype(self, filename):
        griddes = self.cdo.griddes(input=filename)
        griddes = {entry.split('=')[0].strip():entry.split('=')[1].strip() for entry in griddes if '=' in entry}
        return griddes['gridtype']

    def _get_code(self, filename):
        showcode = self.cdo.showcode(input=filename)
        print(showcode)
        return showcode

    @cached
    def _to_regular(self, filename):
        """converts ecmwf spectral grib data to regular gaussian netcdf.

        cdo is used to convert ecmwf grid data to netcdf depending on the gridtype:
        For 'gaussian_reduced': cdo -R
            'spectral'        : cdo sp2gpl

        This follows the recommendation from the ECMWF Era5 Documentation.
        We also invert the latitudes to stick with cmor standard.

        """
        gridtype = self._gridtype(filename)
        if gridtype == 'gaussian_reduced':
            #return self.cdo.copy(options='-R -f nc', input=filename)
            gaussian = self.cdo.setgridtype('regular', options='-f nc', input=filename)
        elif gridtype == 'spectral':
            gaussian = self.cdo.sp2gpl(options='-f nc', input=filename)
        elif gridtype == 'gaussian':
            gaussian =  self.cdo.copy(options='-f nc', input=filename)
        else:
            raise Exception('unknown grid type for conversion to regular grid: {}'.format(gridtype))
        return self.cdo.invertlat(input=gaussian)

    def get_file_by_date(self, datetime):
        """search for a file in the dataframe that should contain the datetime.
        """
        df = self.df[ (datetime >= self.df['startdate']) & (datetime <= self.df['enddate']) ]
        if len(df['path'].values) == 0:
            raise Exception('no files found for {}'.format(datetime))
        elif len(df['path'].values) > 1:
            raise Exception('date selection must be unique')
        else:
            return df['path'].values[0]

    def data_by_date(self, variable, datetime, nc=True):
        return Dataset(self._tmp_file_by_date(datetime, nc)).variables[variable][0]



class Preprocessor():

    def __init__():
        pass



def copy_dataset(src, varname=None, timestep=None, destination=None):
    if varname is None:
        variables = src.variables
    else:
        variables = {varname:src.variables[varname]}
    if destination is None:
        destination = 'copy.nc'
    dst = Dataset(destination,'w')
    # copy attributes
    for name in src.ncattrs():
        dst.setncattr(name, getattr(src, name))
    # copy dimensions
    for name, dimension in src.dimensions.items():
        if timestep and name=='time':
            length = 1
        else:
            length = (len(dimension) if not dimension.isunlimited() else None)
        #dst.createDimension( name, (len(dimension) if not dimension.isunlimited() else None))
        dst.createDimension( name, length)
        # copy all file data except for the excluded
    for name, variable in variables.items():
        if hasattr(variable, '_FillValue'):
            fill_value = getattr(variable, '_FillValue')
        else:
            fill_value = None
        var = dst.createVariable(name, variable.dtype, variable.dimensions, fill_value=fill_value)
        for attr in variable.ncattrs():
            if attr != '_FillValue':
                dst.variables[name].setncattr(attr, getattr(variable, attr))
        if variable.shape:
            if timestep is None or 'time' not in variable.dimensions:
                dst.variables[name][:] = src.variables[name][:]
            else:
                dst.variables[name][0] = src.variables[name][timestep]
    return dst



def get_vc(dataset, flip=False):
    """Reads the vertical hybrid coordinate from a dataset.
    """
    ak_bnds = None
    bk_bnds = None
    for ak_name in ak_valid:
        if ak_name in dataset.variables:
            ak_bnds = dataset.variables[ak_name]
            logging.info('using {} for akgm'.format(ak_name))
    for bk_name in bk_valid:
        if bk_name in dataset.variables:
            bk_bnds = dataset.variables[bk_name]
            logging.info('using {} for bkgm'.format(bk_name))
    if not all([ak_bnds, bk_bnds]):
        logging.error('could not identify vertical coordinate, tried: {}, {}'.format(ak_valid, bk_valid))
        raise Exception('incomplete input dataset')
    nlev = ak_bnds.shape[0]
    ak = np.zeros([nlev+1], dtype=np.float64)
    bk = np.ones([nlev+1], dtype=np.float64)
    if flip:
        ak[:-1] = np.flip(ak_bnds[:,1])
        bk[:-1] = np.flip(bk_bnds[:,1])
    else:
        ak[1:] = np.flip(ak_bnds[:,1])
        bk[1:] = np.flip(bk_bnds[:,1])
    logging.debug('akgm: {}'.format(ak))
    logging.debug('bkgm: {}'.format(bk))
    return ak, bk



def get_dataset(files, ds_type=None):
    if ds_type == 'NC4Dataset':
        return NC4Dataset(files)
    elif ds_type == 'ECMWF':
        return ECMWF(files)



"""this module should help on managing ESGF metadata
"""

import pandas as pd

try:
    from tqdm import tqdm
except:
    def tqdm(x):
        return x


try:
    from pyesgf.search import SearchConnection
except:
    print('pyesgf client is not installed! please install https://github.com/ESGF/esgf-pyclient to make use of this module.')


DKRZ_HOST = 'esgf-data.dkrz.de'
DKRZ_ESG_URL = 'https://{}/esg-search'.format(DKRZ_HOST)

#cordex  output  EUR-11  GERICS  ECMWF-ERAINT  evaluation  r1i1p1  REMO2015  v1  day  tasmax  v20180813
CORDEX_COLUMNS = ['project_id', 'product', 'domain', 'institute', 'driving_model_id', 'experiment_id', 'member', 'model_id', 'version', 'frequency', 'variable', 'date', 'host']
CMIP5_COLUMNS = ['project_id', 'product', 'domain', 'institute', 'driving_model_id', 'experiment_id', 'member', 'model_id', 'version', 'frequency', 'variable', 'date']

COLUMNS  = {'cordex': CORDEX_COLUMNS,
           'cmip5': CMIP5_COLUMNS}


def logon(host=DKRZ_HOST):
    """logon to ESGF Host.
    """
    print('logon to: '.format(host))
    lm = LogonManager()
    lm.logoff()
    lm.logon(hostname=host, interactive=True, bootstrap=True)
    print('logged on: {}'.format(lm.is_logged_on()))
    return lm.is_logged_on()


def connect(url=DKRZ_ESG_URL):
    """establish ESGF search connection.
    """
    return SearchConnection(url, distrib=True)


def context(attrs, conn=None, verbose=False):
    """get an ESGF search context.
    """
    if conn is None:
        ctx = connect().new_context(**attrs)
    else:
        ctx =  conn.new_context(**attrs)
    if verbose: print('Hit Count: {}'.format(ctx.hit_count))
    return ctx


def datasets(attrs, conn=None, verbose=False):
    """returns dataset results
    """
    return context(attrs, conn, verbose).search()


def dataset_dict(attrs, conn=None, verbose=False):
    return {ds.dataset_id: ds for ds in tqdm(datasets(attrs, conn, verbose))}


def get_opendap_urls(attrs, conn=None, agg=False, verbose=False):
    """agg=True will return aggregation urls
    instead of list of file urls...
    """
    ctx = context(attrs, conn, verbose)
    #for facet, item in ctx.facet_counts.items():
    #    if item: print('{} : {}'.format(facet, item))
    results = ctx.search()
    open_dap = {}
    for result in tqdm(results):
        sources = []
        if verbose: print(result.dataset_id)
        if agg:
           ctx_results = result.aggregation_context().search()
        else:
           ctx_results = result.file_context().search()
        for ctx_result in ctx_results:
            sources.append(ctx_result.opendap_url)
        open_dap[result.dataset_id] = sources
    if verbose: print('found {} datasets'.format(len(open_dap.keys())))
    return open_dap


def _split_dataset_id(dataset_id):
    """splits a dataset id into list of values.
    """
    return (dataset_id.split('|')[0].split('.') + [(dataset_id.split('|')[1])])


def _get_columns(columns=None):
    if columns == 'CORDEX':
        return CORDEX_COLUMNS
    elif columns == 'CMIP5':
        return CMIP5_COLUMNS
    elif columns == 'CMIP6':
        return CMIP6_COLUMNS
    else:
        return columns


def to_pandas(results, columns=None):
    """convert a search result to dataframe"""
    data = [_split_dataset_id(res.dataset_id) for res in tqdm(results)]
    return pd.DataFrame(data, columns=_get_columns(columns))





# -*- coding: utf-8 -*-
# flake8: noqa
"""Domain module

This module defines preconfigured CORDEX domain from csv tables. The module
also contains some tools to create a domain dataset from a csv tables or simply
from grid information.

Example:

    To get a list of available implementations, create cordex domains, write
    them to netcdf with some dummy data, you can use ,e.g.,::

        from cordex import domain as dm

        eur11 = dm.cordex_domain('EUR-11')

"""

import numpy as np
import pandas as pd
import xarray as xr

from .tables import domains as TABLES
from . import tables
from . import cf
from . import utils


def domain_from_table(short_name, table):
    """creates domain instance from a pandas dataframe row.
    """
    return Domain(short_name=short_name, **dict(table.loc[short_name]))


def cordex_domain(short_name, dummy=False, **kwargs):
    """Creates an xarray dataset containg the domain grid definitions.

    Parameters
    ----------
    short_name:
        Name of the Cordex Domain.
    dummy : str or logical
        Name of dummy field, if dummy=topo, the cdo topo operator will be
        used to create some dummy topography data. dummy data is useful for
        looking at the domain with ncview.

    Returns
    -------
    Dataset : xarray.core.Dataset
        Dataset containing the coordinates.

    """
    config = pd.concat(TABLES.values()).loc[short_name]
    return create_dataset(**config, dummy=dummy)


def create_dataset(nlon, nlat, dlon, dlat, ll_lon, ll_lat, 
        pollon, pollat, dummy=False,  **kwargs):
    """Create domain dataset from grid information.

    Parameters
    ----------
    nlon : int
        longitudal number of grid boxes
    nlat : int
        latitudal number of grid boxes
    dlon : float
        longitudal resolution (degrees)
    dlat : float
        latitudal resolution (degrees)
    ll_lon : float
        lower left rotated longitude (degrees)
    ll_lat : float
        lower left rotated latitude (degrees)
    pollon : float
        pol longitude (degrees)
    pollat : float
        pol latitude (degrees)
    """
    rlon, rlat = _init_grid(nlon, nlat, dlon, dlat, ll_lon, ll_lat)
    lon, lat = rotated_coord_transform(*_stack(rlon, rlat), pollon, pollat)
    pole = _grid_mapping(pollon, pollat)
    return _get_dataset(rlon, rlat, lon, lat, pole, dummy=dummy)


def _get_dataset(rlon, rlat, lon, lat, pole, dummy=None, mapping_name=None, attrs=None):
    if mapping_name is None:
        mapping_name = cf.DEFAULT_MAPPING_NCVAR
    data_vars={mapping_name: pole}

    ds = xr.Dataset(
        data_vars=data_vars,
        coords=dict(
            rlon=(["rlon"], rlon),
            rlat=(["rlat"], rlat),
            lon=(["rlat", "rlon"], lon),
            lat=(["rlat", "rlon"], lat),
        ),
        attrs=attrs
    )

    for key, coord  in ds.coords.items():
        coord.encoding['_FillValue'] = None
        coord.attrs = cf.coords[key]

    if dummy:
        if dummy is True:
            dummy_name = 'dummy'
        else:
            dummy_name = dummy
        dummy = xr.DataArray(
            data=np.zeros((len(rlat), len(rlon))),
            dims=["rlat", "rlon"],
            coords=dict(
                rlon=(["rlon"], rlon),
                rlat=(["rlat"], rlat),
                lon=(["rlat", "rlon"], lon),
                lat=(["rlat", "rlon"], lat),
            ),
        )
        dummy.attrs = { 'grid_mapping' : mapping_name,
                        'coordinates'  :  'lon lat' }
        ds[dummy_name] = dummy
        if dummy_name == 'topo':
            from cdo import Cdo
            tmp = utils.get_tempfile()
            ds.to_netcdf(tmp)
            topo = Cdo().topo(tmp, returnXDataset=True)['topo'][:]
            ds[dummy_name] = topo

    return ds


def _grid_mapping(pollon, pollat, mapping_name=None):
    """creates a grid mapping DataArray object"""
    if mapping_name is None:
        mapping_name = cf.DEFAULT_MAPPING_NCVAR
    da = xr.DataArray(np.zeros((), dtype=np.int32))
    attrs = cf.mapping.copy()
    attrs['grid_north_pole_longitude'] = pollon
    attrs['grid_north_pole_latitude']  = pollat
    da.attrs = attrs
    return da


def _init_grid(nlon, nlat, dlon, dlat, ll_lon, ll_lat):
    """create coordinate arrays from lower left longitude and latitude"""
    rlon = np.array([ll_lon+i*dlon for i in range(0,nlon)], dtype=np.float64)
    rlat = np.array([ll_lat+i*dlat for i in range(0,nlat)], dtype=np.float64)
    return rlon, rlat


def _stack(x, y):
    """Stack 1d arrays into 2d fields."""
    tmp_x = np.array(x).squeeze()
    tmp_y = np.array(y).squeeze()
    return np.vstack(len(tmp_y)*(tmp_x,)), np.hstack(len(tmp_x)*(tmp_y[:, np.newaxis],))


def rotated_coord_transform(lon, lat, np_lon, np_lat,
                            direction='rot2geo'):
    """Transforms a coordinate into a rotated grid coordinate and vice versa.

    The coordinates have to be given in degree and will be returned in degree.

    Parameters
    ----------
    lon : float array like
        Longitude coordinate.
    lat : float array like
        Latitude coordinate.
    np_lon : float array like
        Longitude coordinate of the rotated north pole.
    np_lat : float array like
        Latitude coordinate of the rotated north pole.
    direction : str
        Direction of the rotation.
        Options are: 'rot2geo' (default) for a transformation to regular
        coordinates from rotated. 'geo2rot' transforms regular coordinates
        to rotated.

    Returns
    -------
    lon_new : array like
        New longitude coordinate.
    lat_new : array like
        New latitude coordinate.
    """

    # Convert degrees to radians
    lon = np.deg2rad(lon)
    lat = np.deg2rad(lat)

    theta = 90. - np_lat # Rotation around y-axis
    phi = np_lon + 180.  # Rotation around z-axis

    # Convert degrees to radians
    phi = np.deg2rad(phi)
    theta = np.deg2rad(theta)

    # Convert from spherical to cartesian coordinates
    x = np.cos(lon) * np.cos(lat)
    y = np.sin(lon) * np.cos(lat)
    z = np.sin(lat)

    # Regular -> Rotated
    if direction == 'geo2rot':

        x_new = (np.cos(theta) * np.cos(phi) * x +
                 np.cos(theta) * np.sin(phi) * y +
                 np.sin(theta) * z)
        y_new = (- np.sin(phi) * x +
                   np.cos(phi) * y)
        z_new = (- np.sin(theta) * np.cos(phi) * x -
                   np.sin(theta) * np.sin(phi) * y +
                   np.cos(theta) * z)

    # Rotated -> Regular
    elif direction == 'rot2geo':

        phi = - phi
        theta = - theta

        x_new = (np.cos(theta) * np.cos(phi) * x +
                 np.sin(phi) * y +
                 np.sin(theta) * np.cos(phi) * z)
        y_new = (- np.cos(theta) * np.sin(phi) * x +
                   np.cos(phi) * y -
                   np.sin(theta) * np.sin(phi) * z)
        z_new = (- np.sin(theta) * x +
                   np.cos(theta) * z)

    # Convert cartesian back to spherical coordinates
    lon_new = np.arctan2(y_new, x_new)
    lat_new = np.arcsin(z_new)

    # Convert radians back to degrees
    lon_new = np.rad2deg(lon_new)
    lat_new = np.rad2deg(lat_new)

    return lon_new, lat_new





#! /usr/bin/python
# coding: utf-8

import tempfile


def get_tempfile():
    """Creates a temporay filename.
    """
    return tempfile.mkstemp()[1]


def copy_dataset(src, varname=None, timestep=None, destination=None):
    """copy an existing NetCDF dataset on disk"""
    if varname is None:
        variables = src.variables
    else:
        variables = {varname:src.variables[varname]}
    if destination is None:
        destination = 'copy.nc'
    dst = Dataset(destination,'w')
    # copy attributes
    for name in src.ncattrs():
        dst.setncattr(name, getattr(src, name))
    # copy dimensions
    for name, dimension in src.dimensions.items():
        if timestep and name=='time':
            length = 1
        else:
            length = (len(dimension) if not dimension.isunlimited() else None)
        #dst.createDimension( name, (len(dimension) if not dimension.isunlimited() else None))
        dst.createDimension( name, length)
        # copy all file data except for the excluded
    for name, variable in variables.items():
        if hasattr(variable, '_FillValue'):
            fill_value = getattr(variable, '_FillValue')
        else:
            fill_value = None
        var = dst.createVariable(name, variable.dtype, variable.dimensions, fill_value=fill_value)
        for attr in variable.ncattrs():
            if attr != '_FillValue':
                dst.variables[name].setncattr(attr, getattr(variable, attr))
        if variable.shape:
            if timestep is None or 'time' not in variable.dimensions:
                dst.variables[name][:] = src.variables[name][:]
            else:
                dst.variables[name][0] = src.variables[name][timestep]
    return dst



"""This module should provide some help to mask data.
"""

import geopandas as gpd
import regionmask

from . import _address as addr

"""convert to WGS84 latitude-longitude projection"""
WGS84 = "EPSG:4326"






def get_geodata(shapefile, to_crs=WGS84, **kwargs):
    shp = gpd.read_file(shapefile)
    if to_crs is not None:
        shp = shp.to_crs(to_crs)
    return shp 


def get_regionmask(geodata, **kwargs):
    return regionmask.from_geopandas(geodata, **kwargs)


# tas_mask = laender.mask(tas, lon_name='lon', lat_name='lat')
# mask_2D = regionmask.mask_geopandas(shp, tas.lon, tas.lat)

def gridded_mask(rmask, da=None, lon=None, lat=None):
    if da is not None:
        if lon is None: lon='lon'
        if lat is None: lat='lat'
        return rmask.mask(da, lon_name=lon, lat_name=lat)
    else:
        if lon is None or lat is None:
            raise Exception('Missing lon or lat attribute for regionmask.mask_geopandas')
        return regionmask.mask_geopandas(rmask, lon, lat)





import xarray as xr



GRID_MAPPING = 'grid_mapping'


def get_grid_mapping(ds, varname=None):
    """Returns the grid mapping variable of a netcdf variable or dataset.
    """
    if varname:
        return get_grid_mapping_from_variable(ds, varname)
    else:
        return get_grid_mapping_from_dataset(ds)
#    if hasattr(da, GRID_MAPPING):


def get_grid_mapping_from_variable(ds, varname):
    """Returns the grid mapping variable associated with a netcdf variable.
    """
    mapping_name = get_grid_mapping_name_from_variable(ds[varname])
    if mapping_name is not None and mapping_name in ds.variables:
        return ds[mapping_name]
    else:
        return None

def get_grid_mapping_from_dataset(ds):
    """Returns the grid mapping variables from a dataset.
    """
    return [mapping for mapping in (get_grid_mapping(ds, var) for var in ds.variables) if mapping is not None]


def get_grid_mapping_name_from_variable(var):
    try:
        return getattr(var, GRID_MAPPING)
    except:
        return None


#def get_grid_mapping_names_from_dataset(ds):
#    return [mapping for mapping in (get_grid_mapping_name_from_variable(ds[var]) for var in ds.variables) if mapping is not None]



# -*- coding: utf-8 -*-
# flake8: noqa

import numpy as np

LAT_NAME = 'lat'
LON_NAME = 'lon'
RLAT_NAME = 'rlat'
RLON_NAME = 'rlon'


coords = \
        {
   'rlon'         : {'axis': 'X',
                     'standard_name': 'grid_longitude',
                     'long_name'    : 'longitude in rotated pole grid',
                     'units'        : 'degrees'},
   'rlat'         : {'axis': 'Y',
                     'standard_name': 'grid_latitude',
                     'long_name'    : 'latitude in rotated pole grid',
                     'units'        : 'degrees'},
   'lon'          : {
                     'standard_name': 'longitude',
                     'long_name'    : 'longitude',
                     'units'        : 'degrees_east'},
   'lat'          : {
                     'standard_name': 'latitude',
                     'long_name'    : 'latitude',
                     'units'        : 'degrees_north'},
   }


mapping = \
     { 'grid_mapping_name' : 'rotated_latitude_longitude',
                     'grid_north_pole_latitude' : 90.,
                     'grid_north_pole_longitude': 180.,
                     'north_pole_grid_longitude': 0.}


DEFAULT_MAPPING_NCVAR = 'rotated_latitude_longitude'
